// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]

package org.xwt;
import java.io.*;
import java.util.*;
import org.xwt.js.*;
import org.xwt.util.*;

/** A ByteStream encapsulates a <i>source</i> of a stream of bytes --
 *  currently either a local file or a Base64-encoded XMLRPC/SOAP element */
public class ByteStream extends JS.Obj {

    private byte[] bytes = null;
    private File file = null;

    private ByteStream() { setSeal(true); }
    ByteStream(String filename) { this(); this.file = new File(filename); }
    ByteStream(byte[] bytes) { this(); this.bytes = bytes; }

    public String toString() { return "ByteStream, source=" + (file == null ? "memory" : file.getAbsolutePath()); }

    public Object get(Object name) {
        if (name.equals("getUTF")) return getUTF;
        else if (name.equals("getDOM")) return getDOM;
        else if (name.equals("fileName")) return file == null ? null : file.getAbsolutePath();
        else return null;
    }

    private class XMLHelper extends XML {
        Vector obStack = new Vector();

        public XMLHelper() { super(BUFFER_SIZE); }

        public void startElement(XML.Element c) throws XML.SchemaException {
            JS o = new JS.Obj();
            o.put("$name", c.localName);
            for(int i=0; i<c.len; i++) o.put(c.keys[i], c.vals[i]);
            o.put("$numchildren", new Integer(0));
            obStack.addElement(o);
        }

        public void endElement(XML.Element c) throws XML.SchemaException {
            if (obStack.size() == 1) return;
            JS me = (JS)obStack.lastElement();
            obStack.setSize(obStack.size() - 1);
            JS parent = (JS)obStack.lastElement();
            int numchildren = ((Integer)parent.get("$numchildren")).intValue();
            parent.put("$numchildren", new Integer(numchildren + 1));
            parent.put(new Integer(numchildren), me);
        }

        public void characters(char[] ch, int start, int length) throws XML.SchemaException {
            String s = new String(ch, start, length);
            JS parent = (JS)obStack.lastElement();
            int numchildren = ((Integer)parent.get("$numchildren")).intValue();
            Object lastChild = parent.get(new Integer(numchildren - 1));
            if (lastChild instanceof String) {
                parent.put(new Integer(numchildren - 1), lastChild + s);
            } else {
                parent.put("$numchildren", new Integer(numchildren + 1));
                parent.put(new Integer(numchildren), s);
            }
        }

        public void whitespace(char[] ch, int start, int length) {}

        public JS doParse() throws JS.Exn {
            try { 
                InputStream is = getInputStream();
                BufferedReader r = new BufferedReader(new InputStreamReader(is));
                parse(r);
            } catch (XML.XMLException e) {
                throw new JS.Exn("error parsing XML: " + e.toString());
            } catch (IOException e) {
                if (Log.on) Log.log(ByteStream.class, "IO Exception while reading from file");
                if (Log.on) Log.log(ByteStream.class, e);
                throw new JS.Exn("error reading from ByteStream");
            }
            return obStack.size() >= 1 ? (JS)obStack.elementAt(0) : null;
        }
    }

    public InputStream getInputStream() throws IOException {
        return file == null ?
            (InputStream)(new ByteArrayInputStream(bytes)) :
            (InputStream)(new FileInputStream(file));
    }

    public void writeTo(OutputStream os) throws IOException {
        InputStream is = getInputStream();
        byte[] buf = new byte[1024];
        while(true) {
            int numread = is.read(buf, 0, 1024);
            if (numread == -1) break;
            if (Log.on) Log.log(this, "wrote " + numread + " bytes");
            os.write(buf, 0, numread);
        }
        os.flush();

        // we have to close this because flush() doesn't work on Win32-GCJ
        os.close();
    }

    private final JS.Callable getDOM = new JS.Callable() {
            public Object call(JS.Array args) throws JS.Exn {
                if (args.length() != 0) return null;
                return new XMLHelper().doParse();
            }
        };

    private final JS.Callable getUTF = new JS.Callable() {
            public Object call(JS.Array args) throws JS.Exn {
                if (args.length() != 0) return null;
                try {
                    CharArrayWriter caw = new CharArrayWriter();
                    InputStream is = getInputStream();
                    BufferedReader r = new BufferedReader(new InputStreamReader(is));
                    char[] buf = new char[1024];
                    while(true) {
                        int numread = r.read(buf, 0, 1024);
                        if (numread == -1) break;
                        caw.write(buf, 0, numread);
                    }
                    return caw.toString();
                } catch (IOException e) {
                    if (Log.on) Log.log(ByteStream.class, "IO Exception while reading from file");
                    if (Log.on) Log.log(ByteStream.class, e);
                    throw new JS.Exn("error while reading from ByteStream");
                }
            }
        };

}

