// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import org.bouncycastle.util.encoders.Base64;
import org.xwt.util.*;
import java.io.*;
import java.util.*;

/** 
 *  A Surface, as described in the XWT Reference.
 *
 *  Platform subclasses should include an inner class subclass of
 *  Surface to return from the Platform._createSurface() method
 *
 *  Note that the members in the section 'state variables' are either
 *  in real-time (the actual size/position/state), or in
 *  MessageQueue-time (the size/position/state at the time that the
 *  now-executing message was enqueued). This distinction is important.
 */
public abstract class Surface {

    // Static Data ////////////////////////////////////////////////////////////////////////////////

    /** true iff a user-created surface was created */
    static boolean refreshableSurfaceWasCreated = false;

    /** a reference to the most recently enqueued Move message; used to throttle the message rate */
    private static Message lastMoveMessage = null;

    /** all instances of Surface which need to be refreshed by the MessageQueue */
    public static Vec allSurfaces = new Vec();

    /** true iff the alt button is pressed down, in real time */
    public static boolean alt = false;
    
    /** true iff the control button is pressed down, in real time */
    public static boolean control = false;

    /** true iff the shift button is pressed down, in real time */
    public static boolean shift = false;

    /** true iff button 1 is depressed, in MessageQueue-time */
    public static boolean button1 = false;

    /** true iff button 2 is depressed, in MessageQueue-time */
    public static boolean button2 = false;

    /** true iff button 3 is depressed, in MessageQueue-time */
    public static boolean button3 = false;

    /** true iff all surfaces created from now on should be scarred */
    public static boolean scarAllSurfacesFromNowOn = false;

    /** false if the surface has never been rendered; used to determine if the surface should be repositioned to be centered on the screen */
    public boolean centerSurfaceOnRender = true;

    /** the x position of the mouse, relative to this Surface, in MessageQueue-time */
    public int mousex;

    /** the y position of the mouse, relative to this Surface, in MessageQueue-time */
    public int mousey;

    /** True iff this surface is minimized, in real time */
    public boolean minimized = false;

    /** True iff this surface is maximized, in real time */
    public boolean maximized = false;

    /** The name of the cursor on this surface -- this value fluctuates during rendering, so it may not be accurate;
     *  syncCursor() is called once the value is stable, to prevent the "flickering cursor" phenomenon
     */
    public String cursor = "default";

    /** The Box at the root of this surface */
    public Box root;

    /** The number of SizeChange/PosChange traps triggered since the last successful render -- used to detect infinite loops */
    public int sizePosChangesSinceLastRender = 0;


    // Used For Simulating Clicks and DoubleClicks /////////////////////////////////////////////////

    /** the x-position of the mouse the last time a Press message was enqueued */
    int last_press_x = Integer.MAX_VALUE;

    /** the y-position of the mouse the last time a Press message was enqueued */
    int last_press_y = Integer.MAX_VALUE;

    /** the last button to recieve a Click message; used for simulating DoubleClick's */
    static int lastClickButton = 0;

    /** the last time a Click message was processed; used for simulating DoubleClick's */
    static long lastClickTime = 0;
    
    
    // Methods to be overridden by subclasses ///////////////////////////////////////////////////////

    /** when this method is invoked, the surface should push itself to the back of the stacking order */
    public abstract void toBack();

    /** when this method is invoked, the surface should pull itself to the front of the stacking order */
    public abstract void toFront();

    /** sets the <i>actual</i> cursor for this surface to the cursor referenced by <tt>cursor</tt> */
    public abstract void syncCursor();

    /** If <tt>b == true</tt>, make the window invisible; otherwise, make it non-invisible. */
    public abstract void setInvisible(boolean b);

    /** If <tt>b == true</tt>, maximize the surface; otherwise, un-maximize it. */
    protected abstract void _setMaximized(boolean b);

    /** If <tt>b == true</tt>, minimize the surface; otherwise, un-minimize it. */
    protected abstract void _setMinimized(boolean b);

    /** Sets the surface's width and height. */
    protected abstract void setSize(int width, int height);

    /** Sets the surface's x and y position. */
    public abstract void setLocation(int x, int y);

    /** Sets the surface's title bar text, if applicable */
    public abstract void setTitleBarText(String s);

    /** Sets the surface's title bar text, if applicable */
    public abstract void setIcon(Picture i);

    /** copies a region from the doublebuffer to this surface */
    public abstract void blit(DoubleBuffer source, int sx, int sy, int dx, int dy, int dx2, int dy2);

    /** Destroy the surface */
    public abstract void _dispose();

    /** Notifies the surface that limits have been imposed on the surface's size */
    public void setLimits(int min_width, int min_height, int max_width, int max_height) { }


    // Helper methods for subclasses ////////////////////////////////////////////////////////////

    protected final void Press(final int button) {
        last_press_x = mousex;
        last_press_y = mousey;

        if (button == 1) button1 = true;
        else if (button == 2) button2 = true;
        else if (button == 3) button3 = true;

        if (button == 1) new SimpleMessage("Press1", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 2) new SimpleMessage("Press2", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 3) {
            final Box who = Box.whoIs(root, mousex, mousey);
            MessageQueue.add(new Message() { public void perform() {
                Platform.clipboardReadEnabled = true;
                root.put("Press3", Boolean.TRUE);
                Platform.clipboardReadEnabled = false;
            }});
        }
    }

    protected final void Release(int button) {
        if (button == 1) button1 = false;
        else if (button == 2) button2 = false;
        else if (button == 3) button3 = false;

        if (button == 1) new SimpleMessage("Release1", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 2) new SimpleMessage("Release2", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 3) new SimpleMessage("Release3", Boolean.TRUE, Box.whoIs(root, mousex, mousey));

        if (Platform.needsAutoClick() && Math.abs(last_press_x - mousex) < 5 && Math.abs(last_press_y - mousey) < 5) Click(button);
        last_press_x = Integer.MAX_VALUE;
        last_press_y = Integer.MAX_VALUE;
    }

    protected final void Click(int button) {
        if (button == 1) new SimpleMessage("Click1", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 2) new SimpleMessage("Click2", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 3) new SimpleMessage("Click3", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        if (Platform.needsAutoDoubleClick()) {
            long now = System.currentTimeMillis();
            if (lastClickButton == button && now - lastClickTime < 350) DoubleClick(button);
            lastClickButton = button;
            lastClickTime = now;
        }
    }

    protected final void DoubleClick(int button) {
        if (button == 1) new SimpleMessage("DoubleClick1", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 2) new SimpleMessage("DoubleClick2", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
        else if (button == 3) new SimpleMessage("DoubleClick3", Boolean.TRUE, Box.whoIs(root, mousex, mousey));
    }

    /** sends a KeyPressed message; subclasses should not add the C- or A- prefixes, nor should they capitalize alphabet characters */
    protected final void KeyPressed(String key) {
        if (key == null) return;

        if (key.toLowerCase().endsWith("shift")) shift = true;
        else if (shift) key = key.toUpperCase();

        if (key.toLowerCase().equals("alt")) alt = true;
        else if (alt) key = "A-" + key;

        if (key.toLowerCase().endsWith("control")) control = true;
        else if (control) key = "C-" + key;

        final String fkey = key;
        MessageQueue.add(new KMessage(key));
    }

    // This is implemented as a private static class instead of an anonymous class to work around a GCJ bug
    private class KMessage implements Message {
        String key = null;
        public KMessage(String k) { key = k; }
        public void perform() {
            if (key.equals("C-v") || key.equals("A-v")) Platform.clipboardReadEnabled = true;
            outer: for(int i=0; i<keywatchers.size(); i++) {
                Box b = (Box)keywatchers.elementAt(i);
                for(Box cur = b; cur != null; cur = cur.getParent())
                    if (cur.invisible) continue outer;
                b.put("KeyPressed", key);
            }
            Platform.clipboardReadEnabled = false;
        }
    }

    /** sends a KeyReleased message; subclasses should not add the C- or A- prefixes, nor should they capitalize alphabet characters */
    protected final void KeyReleased(final String key) {
        if (key == null) return;
        if (key.toLowerCase().equals("alt")) alt = false;
        else if (key.toLowerCase().equals("control")) control = false;
        else if (key.toLowerCase().equals("shift")) shift = false;
        MessageQueue.add(new Message() { public void perform() {
            outer: for(int i=0; i<keywatchers.size(); i++) {
                Box b = (Box)keywatchers.elementAt(i);
                for(Box cur = b; cur != null; cur = cur.getParent())
                    if (cur.invisible) continue outer;
                b.put("KeyReleased", key);
            }
        }});
    }

    /**
     *  Notify XWT that the mouse has moved. If the mouse leaves the
     *  surface, but the host windowing system does not provide its new
     *  position (for example, a Java MouseListener.mouseExited()
     *  message), the subclass should use (-1,-1).
     */
    protected final void Move(final int newmousex, final int newmousey) {
        MessageQueue.add(lastMoveMessage = new Message() { public void perform() {
            synchronized(Surface.this) {

                // if move messages are arriving faster than we can process them, we just start ignoring them
                if (lastMoveMessage != this) return;

                int oldmousex = mousex;
                int oldmousey = mousey;
                mousex = newmousex;
                mousey = newmousey;

                String oldcursor = cursor;
                cursor = "default";

                // Root gets motion events outside itself (if trapped, of course)
                if (!root.inside(oldmousex, oldmousey) && !root.inside(mousex, mousey) && (button1 || button2 || button3))
                    root.put("Move", Boolean.TRUE);

                root.Move(oldmousex, oldmousey, mousex, mousey);
                if (!cursor.equals(oldcursor)) syncCursor();
            }
        }});
    }

    protected final void SizeChange(final int width, final int height) {
        MessageQueue.add(new Message() { public void perform() {
            if (width == root.width && height == root.height) return;
            root.width = width;
            root.height = height;
            root.needs_reflow = true;
        }});
        abort = true;
    }

    protected final void PosChange(final int x, final int y) {
        MessageQueue.add(new Message() { public void perform() {
            root.put("x", new Integer(x));
            root.put("y", new Integer(y));
        }});
    }

    protected final void Close() { new SimpleMessage("Close", Boolean.TRUE, root); }
    protected final void Minimized(boolean b) { minimized = b; new SimpleMessage("Minimized", b ? Boolean.TRUE : Boolean.FALSE, root); }
    protected final void Maximized(boolean b) { maximized = b; new SimpleMessage("Maximized", b ? Boolean.TRUE : Boolean.FALSE, root); }
    protected final void Focused(boolean b) { new SimpleMessage("Focused", b ? Boolean.TRUE : Boolean.FALSE, root); }
    public static void Refresh() { MessageQueue.refresh(); }

    // the following value is split into two int's to work around GCJ bug java/6393

    /** This is how subclasses signal a 'shallow dirty', indicating that although the backbuffer is valid, the screen is not */
    public final void Dirty(int x, int y, int w, int h) {
        screenDirtyRegions.dirty(x, y, w, h);
        Refresh();
    }


    // Private Instance Data /////////////////////////////////////////////////////////////////////////////////////////////

    /** The automatic double buffer for the root box */
    DoubleBuffer backbuffer = null;

    /** Dirty regions on the backbuffer which need to be rebuilt using Box.render() */
    private DirtyList backbufferDirtyRegions = new DirtyList();

    /** Dirty regions on the screen which need to be rebuilt using Surface.blit() */
    private DirtyList screenDirtyRegions = new DirtyList();

    /** A list of all the Boxes on this Surface that should be notified of keyboard events */
    Vec keywatchers = new Vec();

    /** When set to true, render() should abort as soon as possible and restart the rendering process */
    static volatile boolean abort = false;

    /** a solid red 10x10 double buffer */
    private DoubleBuffer showRenderBuf = null;

    /** a striped 100x100 double buffer */
    private DoubleBuffer showRenderBuf2 = null;

    /** true iff this window should be scarred */
    private boolean scarred = true;


    // Other Methods ///////////////////////////////////////////////////////////////////////////////

    /** If <tt>b == true</tt>, maximize the surface; otherwise, un-maximize it. */
    public final void setMaximized(boolean b) {
        if (b == maximized) return;
        _setMaximized(b);
        maximized = b;
    }

    /** If <tt>b == true</tt>, minimize the surface; otherwise, un-minimize it. */
    public final void setMinimized(boolean b) {
        if (b == minimized) return;
        _setMinimized(b);
        minimized = b;
    }

    /** wrapper for setSize() which makes sure to dirty the place where the scar used to be */
    void setSize() {
        if (scarred) {
            root.width = Math.max(root.width, scarPicture.getWidth());
            root.height = Math.max(root.height, scarPicture.getHeight());
            dirty(hscar,
                  root.height - vscar - scarPicture.getHeight(),
                  scarPicture.getWidth(),
                  scarPicture.getHeight());
        }
        setSize(root.width, root.height);
    }

    /** Indicates that the Surface is no longer needed */
    public final void dispose(boolean quitIfAllSurfacesGone) {
        if (root == null) return;
        if (Log.on) Log.log(this, "disposing " + this);
        allSurfaces.removeElement(this);
        _dispose();

        // quit when all windows are closed
        if (allSurfaces.size() == 0 && quitIfAllSurfacesGone && Main.doneInitializing) {
            if (Log.on) {
                if (refreshableSurfaceWasCreated) Log.log(this, "exiting because last remaining surface was disposed");
                else Log.log(this, "exiting because no surface was ever created");
            }
            Platform.exit();
        }
    }

    /** Indicates that the backbuffer region x,y,w,h is no longer correct and must be regenerated */
    public void dirty(int x, int y, int w, int h) {
        backbufferDirtyRegions.dirty(x, y, w, h);
        Refresh();
    }

    public Surface(Box root) {
        this.scarred = scarAllSurfacesFromNowOn;
        scarAllSurfacesFromNowOn = true;
        this.root = root;
        if (root.surface != null && root.surface.root == root) {
            root.surface.dispose(false);
        } else {
            root.remove();
        }
        root.surface = this;

        // make sure the root is properly sized
        do {
            abort = false;
            root.reflow();
        } while(abort);

        // this is a bit dangerous since we're passing ourselves to another method before subclasses' ctors have run...        
        backbuffer = Platform.createDoubleBuffer(Platform.getScreenWidth(), Platform.getScreenHeight(), this);

        root.dirty();
        Refresh();
    }

    /** runs the prerender() and render() pipelines in the root Box to regenerate the backbuffer, then blits it to the screen */
    public synchronized void render() {

        // make sure the root is properly sized
        do {
            abort = false;
            root.reflow();
            // update mouseinside and trigger Enter/Leave as a result of box size/position changes
            String oldcursor = cursor;
            cursor = "default";
            root.Move(mousex, mousey, mousex, mousey);
            if (!cursor.equals(oldcursor)) syncCursor();
        } while(abort);

        if (centerSurfaceOnRender) {
            centerSurfaceOnRender = false;
            int x = (Platform.getScreenWidth() - root.width) / 2;
            int y = (Platform.getScreenHeight() - root.height) / 2;
            setLocation(x, y);
            root.x = x;
            root.y = y;
        }

        sizePosChangesSinceLastRender = 0;
        int[][] dirt = backbufferDirtyRegions.flush();
        for(int i = 0; dirt != null && i < dirt.length; i++) {
            if (dirt[i] == null) continue;
            int x = dirt[i][0];
            int y = dirt[i][1];
            int w = dirt[i][2];
            int h = dirt[i][3];
            if (x < 0) x = 0;
            if (y < 0) y = 0;
            if (x+w > root.width) w = root.width - x;
            if (y+h > root.height) h = root.height - y;
            if (w <= 0 || h <= 0) continue;

            root.render(0, 0, x, y, w, h, backbuffer);
            
            // if any area under the scar was repainted, rescar that area
            if (scarred && x < hscar + scarPicture.getWidth() &&
                y + h > root.height - scarPicture.getHeight() - vscar) {
                int _x1 = Math.max(x, hscar);
                int _x2 = Math.min(x + w, hscar + scarPicture.getWidth());
                int _y1 = Math.max(y, root.height - scarPicture.getHeight() - vscar);
                int _y2 = Math.min(y + h, root.height - vscar);
                
                backbuffer.drawPicture(scarPicture, _x1, _y1, _x2, _y2,
                                       _x1 - (hscar),
                                       _y1 - (root.height - scarPicture.getHeight() - vscar),
                                       _x2 - (hscar),
                                       _y2 - (root.height - scarPicture.getHeight() - vscar)
                                       );
            }

            if (abort) {

                // x,y,w,h is only partially reconstructed, so we must be careful not to re-blit it
                blitDirtyScreenRegions(x, y, w, h);
                screenDirtyRegions.dirty(x, y, w, h);

                // put back all the dirty regions we haven't yet processed (including the current one)
                for(int j=i; j<dirt.length; j++)
                    if (dirt[j] != null)
                        backbufferDirtyRegions.dirty(dirt[j][0], dirt[j][1], dirt[j][2], dirt[j][3]);

                // tail-recurse
                render();
                return;
            }

            // now that we've reconstructed this region in the backbuffer, queue it to be reblitted
            screenDirtyRegions.dirty(x, y, w, h);
        }

        // blit out all the areas we've just reconstructed
        blitDirtyScreenRegions();
    }

    /** blits from the backbuffer to the screen for all regions of the screen which have become dirty */
    public synchronized void blitDirtyScreenRegions() { blitDirtyScreenRegions(-1, -1, 0, 0); }

    /** same as blitDirtyScreenRegions(), except that it will skip any regions within a,b,c,d */
    private synchronized void blitDirtyScreenRegions(int a, int b, int c, int d) {

        int[][] dirt = screenDirtyRegions.flush();
        if (Main.showRenders && dirt != null && dirt.length > 0 && a == -1)
            blit(backbuffer, 0, 0, 0, 0, root.width, root.height);

        for(int i = 0; dirt != null && i < dirt.length; i++) {
            if (dirt[i] == null) continue;
            int x = dirt[i][0];
            int y = dirt[i][1];
            int w = dirt[i][2];
            int h = dirt[i][3];
            if (x < 0) x = 0;
            if (y < 0) y = 0;
            if (x+w > root.width) w = root.width - x;
            if (y+h > root.height) h = root.height - y;
            if (w <= 0 || h <= 0) continue;

            // if any part of this region falls within the "bad region", just skip it
            boolean hhit = (x >= a && x <= a + c) || (x+w >= a && x+w <= a + c);
            boolean vhit = (y >= b && y <= b + d) || (y+h >= b && y+h <= b + d);
            if (hhit && vhit) {
                screenDirtyRegions.dirty(x, y, w, h);
                continue;
            }

            blit(backbuffer, x, y, x, y, w + x, h + y);
            
            if (Main.showRenders) {
                if (showRenderBuf == null) {
                    showRenderBuf = Platform.createDoubleBuffer(10, 10, this);
                    showRenderBuf.fillRect(0, 0, 10, 10, 0x00FF0000);
                    showRenderBuf2 = Platform.createDoubleBuffer(100, 100, this);
                    for(int y1 = 0; y1<100; y1++)
                        for(int x1 = 0; x1<100; x1++)
                            if ((x1 + y1) % 5 == 0)
                                showRenderBuf2.fillRect(x1, y1, x1 + 1, y1 + 1, 0x00FF0000);
                }
                for(int x1 = x; x1<x + w; x1 += 100)
                    for(int y1 = y; y1< y + h; y1 += 100) {
                        blit(showRenderBuf2, 0, 0, x1, y1, Math.min(x1 + 100, x + w), Math.min(y1 + 100, y + h));
                    }
                for(int j=x; j<x + w; j += 10) {
                    blit(showRenderBuf, 0, 0, j, y, Math.min(j+ 10, x + w), y + 1);
                    blit(showRenderBuf, 0, 0, j, y + h, Math.min(j + 10, x + w), y + h + 1);
                }
                for(int j=y; j<y + h; j += 10) {
                    blit(showRenderBuf, 0, 0, x, j, x + 1, Math.min(j + 10, y + h));
                    blit(showRenderBuf, 0, 0, x + w, j, x + w + 1, Math.min(j + 10, y + h));
                }
            }

        }
    }

    // FEATURE: reinstate recycler
    public class SimpleMessage implements Message {
        
        private Box boxContainingMouse;
        private Object value;
        public String name;
        
        SimpleMessage(String name, Object value, Box boxContainingMouse) {
            this.boxContainingMouse = boxContainingMouse;
            this.name = name;
            this.value = value;
            MessageQueue.add(this);
        }
        
        public void perform() { boxContainingMouse.put(name, value); }
        public String toString() { return "SimpleMessage [name=" + name + ", value=" + value + "]"; }
        
    }

    // Scar-Related Stuff ////////////////////////////////////////////////////////////////////

    /** The scar's horizontal offset */
    int hscar = 0;

    /** The scar's vertical offset */
    int vscar = 0;

    /** the scar image drawn on the bottom right hand corner */
    static Picture scarPicture = null;

}
