// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt.plat;

import org.xwt.*;
import org.xwt.util.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.image.*;
import java.awt.event.*;

/** Platform subclass for all VM's providing AWT 1.1 functionality */
public class AWT extends Platform {

    protected String getDescriptiveName() { return "Generic JDK 1.1+ with AWT"; }
    protected DoubleBuffer _createDoubleBuffer(int w, int h, Surface owner) { return new AWTDoubleBuffer(w, h); }
    protected Picture _createPicture(int[] b, int w, int h) { return new AWTPicture(b, w, h); }
    protected int _getScreenWidth() { return Toolkit.getDefaultToolkit().getScreenSize().width; }
    protected int _getScreenHeight() { return Toolkit.getDefaultToolkit().getScreenSize().height; }
    protected Surface _createSurface(Box b, boolean framed) { return new AWTSurface(b, framed); }
    protected int _stringWidth(String font, String text) { return getFont(font).metrics.stringWidth(text); }
    protected int _getMaxAscent(String font) { return getFont(font).metrics.getMaxAscent(); }
    protected int _getMaxDescent(String font) { return getFont(font).metrics.getMaxDescent(); }
    protected boolean _supressDirtyOnResize() { return false; }

    protected void postInit() {
        if (Log.on) Log.log(Platform.class, "               color depth = " + Toolkit.getDefaultToolkit().getColorModel().getPixelSize() + "bpp");
    }

    protected void _criticalAbort(String message) {
        if (Log.on) Log.log(this, message);
        final Dialog d = new Dialog(new Frame(), "XWT Cannot Continue");
        d.setLayout(new BorderLayout());
        TextArea ta = new TextArea("XWT cannot continue because:\n\n" + message, 10, 80);
        ta.setEditable(false);
        d.add(ta, "Center");
        Button b = new Button("OK");
        b.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    d.dispose();
                }
            });
        d.add(b, "South");
        d.setModal(true);
        d.pack();
        d.show();
        new Semaphore().block();
    }

    protected String _getClipBoard() {
        Clipboard cb = Toolkit.getDefaultToolkit().getSystemClipboard();
        if (cb == null) return null;
        Transferable clipdata = cb.getContents(null);
        try { return (String)clipdata.getTransferData(DataFlavor.stringFlavor); } catch (Exception ex) { return null; }
    }

    protected void _setClipBoard(String s) {
        Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
        if (clipboard == null) return;
        StringSelection clipString = new StringSelection(s);
        clipboard.setContents(clipString, clipString);
    }

    /** some platforms (cough, cough, NetscapeVM) have totally broken modifier masks; they will need to override this */
    protected static int modifiersToButtonNumber(int modifiers) {
        if ((modifiers & InputEvent.BUTTON1_MASK) == InputEvent.BUTTON1_MASK) return 1;
        if ((modifiers & InputEvent.BUTTON2_MASK) == InputEvent.BUTTON2_MASK) {
            // ugh, MacOSX reports the right mouse button as BUTTON2_MASK...
            if (System.getProperty("os.name", "").startsWith("Mac OS X")) return 2;
            return 3;
        }
        if ((modifiers & InputEvent.BUTTON3_MASK) == InputEvent.BUTTON3_MASK) {
            // ugh, MacOSX reports the right mouse button as BUTTON2_MASK...
            if (System.getProperty("os.name", "").startsWith("Mac OS X")) return 3;
            return 2;
        }
        return 0;
    }

    static class FileDialogHelper extends FileDialog implements WindowListener, ComponentListener {
        Semaphore s;
        public FileDialogHelper(String suggestedFileName, Semaphore s, boolean write) {
            super(new Frame(), write ? "Save" : "Open", write ? FileDialog.SAVE : FileDialog.LOAD);
            this.s = s;
            addWindowListener(this);
            addComponentListener(this);
            if (suggestedFileName.indexOf(File.separatorChar) == -1) {
                setFile(suggestedFileName);
            } else {
                setDirectory(suggestedFileName.substring(0, suggestedFileName.lastIndexOf(File.separatorChar)));
                setFile(suggestedFileName.substring(suggestedFileName.lastIndexOf(File.separatorChar) + 1));
            }
            show();
        }
        public void windowActivated(WindowEvent e) { }
        public void windowClosed(WindowEvent e) { s.release(); }
        public void windowClosing(WindowEvent e) { }
        public void windowDeactivated(WindowEvent e) { }
        public void windowDeiconified(WindowEvent e) { }
        public void windowIconified(WindowEvent e) { }
        public void windowOpened(WindowEvent e) { }
        public void componentHidden(ComponentEvent e) { s.release(); }
        public void componentMoved(ComponentEvent e) { }
        public void componentResized(ComponentEvent e) { }
        public void componentShown(ComponentEvent e) { }
    };

    protected String _fileDialog(String suggestedFileName, boolean write) {
        final Semaphore s = new Semaphore();
        FileDialogHelper fd = new FileDialogHelper(suggestedFileName, s, write);
        s.block();
        return fd.getDirectory() + File.separatorChar + fd.getFile();
    }


    // Inner Classes /////////////////////////////////////////////////////////////////////////////////////

    protected static class AWTPicture extends Picture {
        public int getHeight() { return i.getHeight(null); }
        public int getWidth() { return i.getWidth(null); } 
        public int[] getData() { return data; }

        int[] data = null;
        public Image i = null;
        private static ColorModel cmodel = new DirectColorModel(32, 0x00FF0000, 0x0000FF00, 0x000000FF, 0xFF000000);
        
        public AWTPicture(int[] b, int w, int h) {
            data = b;
            Image img = Toolkit.getDefaultToolkit().createImage(new MemoryImageSource(w, h, cmodel, b, 0, w));
            MediaTracker mediatracker = new MediaTracker(new Canvas());
            mediatracker.addImage(img, 1);
            try { mediatracker.waitForAll(); } catch (InterruptedException e) { }
            mediatracker.removeImage(img);
            this.i = img;
        }
    }
    
    protected static class AWTDoubleBuffer extends DoubleBuffer {
        
        protected Image i = null;
        protected Graphics g = null;
        
        /** JDK1.1 platforms require that a component be associated with each off-screen buffer */
        static Component component = null;

        protected AWTDoubleBuffer() { }
        public AWTDoubleBuffer(int w, int h) {
            synchronized(AWTDoubleBuffer.class) {
                if (component == null) {
                    component = new Frame();
                    component.setVisible(false);
                    component.addNotify();
                }
            }
            i = component.createImage(w, h);
            g = i.getGraphics();
        }
        
        public int getHeight() { return i == null ? 0 : i.getHeight(null); }
        public int getWidth() { return i == null ? 0 : i.getWidth(null); }
        public void setClip(int x, int y, int x2, int y2) { g.setClip(x, y, x2 - x, y2 - y); }

        public void drawPicture(Picture source, int x, int y) {
            drawPicture(source, x, y, x + source.getWidth(), y + source.getHeight(), 0, 0, source.getWidth(), source.getHeight());
        }

        public void drawPicture(Picture source, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2) {
            g.drawImage(((AWTPicture)source).i, dx1, dy1, dx2, dy2, sx1, sy1, sx2, sy2, null);
        }
        
        public void drawString(String font, String text, int x, int y, int argb) {
            // FEATURE: use an LRU cache for Color objects
            g.setColor(new Color((argb & 0x00FF0000) >> 16, (argb & 0x0000FF00) >> 8, (argb & 0x000000FF)));
            g.setFont(getFont(font));
            g.drawString(text, x, y + 2);
        }
        
        public void fillRect(int x, int y, int x2, int y2, int argb) {
            // FEATURE: use an LRU cache for Color objects
            g.setColor(new Color((argb & 0x00FF0000) >> 16, (argb & 0x0000FF00) >> 8, (argb & 0x000000FF)));
            g.fillRect(x, y, x2 - x, y2 - y);
        }

    }
    
    
    protected static class AWTSurface extends Surface
        implements MouseListener, MouseMotionListener, KeyListener, ComponentListener, WindowListener {

        public void blit(DoubleBuffer s, int sx, int sy, int dx, int dy, int dx2, int dy2) {
            if (ourGraphics == null) ourGraphics = window.getGraphics();
            ourGraphics.drawImage(((AWTDoubleBuffer)s).i, dx + insets.left, dy + insets.top, dx2 + insets.left, dy2 + insets.top,
                                  sx, sy, sx + (dx2 - dx), sy + (dy2 - dy), null);
        }
        
        /** if (component instanceof Frame) then frame == window else frame == null */
        Frame frame = null;
        Window window = null;
        
        /** our component's insets */
        protected Insets insets = new Insets(0, 0, 0, 0);
        
        /** a Graphics context on <code>window</code> */
        protected Graphics ourGraphics = null;
        
        /** some JDKs let us recycle a single Dimension object when calling getSize() */
        Dimension singleSize = new Dimension();
        
        public void toBack() { if (window != null) window.toBack(); }
        public void toFront() { if (window != null) window.toFront(); }
        public void setLocation(int x, int y) { window.setLocation(x, y); }
        public void setTitleBarText(String s) { if (frame != null) frame.setTitle(s); }
        public void setIcon(Picture i) { if (frame != null) frame.setIconImage(((AWTPicture)i).i); }
        public void setSize(int width, int height) { window.setSize(width + (insets.left + insets.right), height + (insets.top + insets.bottom)); }
        public void setInvisible(boolean b) { window.setVisible(!b); }
        protected void _setMinimized(boolean b) { if (Log.on) Log.log(this, "JDK 1.1 platforms cannot minimize or unminimize windows"); }
        protected void _setMaximized(boolean b) {
            if (!b) {
                if (Log.on) Log.log(this, "JDK 1.1 platforms cannot unmaximize windows");
                return;
            }
            window.setLocation(new Point(0, 0));
            window.setSize(Toolkit.getDefaultToolkit().getScreenSize());
        }

        class InnerFrame extends Frame {
            public InnerFrame() throws java.lang.UnsupportedOperationException { }
            public void update(Graphics gr) { paint(gr); }
            public void paint(Graphics gr) {
                Rectangle r = gr.getClipBounds();

                // ugly hack for Java1.4 dynamicLayout on Win32 -- this catches expansions during smooth resize
                int newwidth = Math.max(r.x - insets.left + r.width, root.width);
                int newheight = Math.max(r.y - insets.top + r.height, root.height);
                if (newwidth > root.width || newheight > root.height)
                    componentResized(window.getWidth() - insets.left - insets.right, window.getHeight() - insets.top - insets.bottom);

                Dirty(r.x - insets.left, r.y - insets.top, r.width, r.height);
            }
        }

        class InnerWindow extends Window {
            public InnerWindow() throws java.lang.UnsupportedOperationException { super(new Frame()); }
            public void update(Graphics gr) { paint(gr); }
            public void paint(Graphics gr) {
                Rectangle r = gr.getClipBounds();
                Dirty(r.x - insets.left, r.y - insets.top, r.width, r.height);
            }
        }

        AWTSurface(Box root, boolean framed) {
            super(root);
            try {
                if (framed) window = frame = new InnerFrame();
                else window = new InnerWindow();

            // this is here to catch HeadlessException on jdk1.4
            } catch (java.lang.UnsupportedOperationException e) {
                if (Log.on) Log.log(this, "Exception thrown in AWTSurface$InnerFrame() -- this should never happen");
                if (Log.on) Log.log(this, e);
            }

            insets = window.getInsets();
            
            window.addMouseListener(this);
            window.addKeyListener(this);
            window.addComponentListener(this);
            window.addMouseMotionListener(this);
            window.addWindowListener(this);

            // IMPORTANT: this must be called before render() to ensure
            // that our peer has been created
            makeVisible();
        }

        protected void makeVisible() { window.setVisible(true); }
        
        public void _dispose() {
            window.removeMouseListener(this);

            // removed to work around a jdk1.3 bug
            /* window.removeKeyListener(this); */

            window.removeComponentListener(this);
            window.removeMouseMotionListener(this);
            window.removeWindowListener(this);
            window.dispose();
        }

        public void syncCursor() {
            if (cursor.equals("crosshair")) window.setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
            else if (cursor.equals("east")) window.setCursor(Cursor.getPredefinedCursor(Cursor.E_RESIZE_CURSOR));
            else if (cursor.equals("move")) window.setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
            else if (cursor.equals("north")) window.setCursor(Cursor.getPredefinedCursor(Cursor.N_RESIZE_CURSOR));
            else if (cursor.equals("northeast")) window.setCursor(Cursor.getPredefinedCursor(Cursor.NE_RESIZE_CURSOR));
            else if (cursor.equals("northwest")) window.setCursor(Cursor.getPredefinedCursor(Cursor.NW_RESIZE_CURSOR));
            else if (cursor.equals("south")) window.setCursor(Cursor.getPredefinedCursor(Cursor.S_RESIZE_CURSOR));
            else if (cursor.equals("southeast")) window.setCursor(Cursor.getPredefinedCursor(Cursor.SE_RESIZE_CURSOR));
            else if (cursor.equals("southwest")) window.setCursor(Cursor.getPredefinedCursor(Cursor.SW_RESIZE_CURSOR));
            else if (cursor.equals("text")) window.setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
            else if (cursor.equals("west")) window.setCursor(Cursor.getPredefinedCursor(Cursor.W_RESIZE_CURSOR));
            else if (cursor.equals("wait")) window.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
            else if (cursor.equals("hand")) window.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
            else window.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }
        
        // AWT Message translation ////////////////////////////////////////////////////////////////
        
        // these functions are all executed in the AWT thread, not the
        // MessageQueue thread. As a result, they must be *extremely*
        // careful about invoking methods on instances of Box. Currently,
        // they should only enqueue messages, use Box.whoIs()
        // (unsynchronized but thought to be safe), and modify members of
        // Surface.
        
        public void componentHidden(ComponentEvent e) { }
        public void componentShown(ComponentEvent e) { }
        public void windowOpened(WindowEvent e) { }
        public void windowClosed(WindowEvent e) { }
        public void windowClosing(WindowEvent e) { Close(); }
        public void windowIconified(WindowEvent e) { Minimized(true); }
        public void windowDeiconified(WindowEvent e) { dirty(0, 0, root.width, root.height); Minimized(false); }
        public void windowActivated(WindowEvent e) { Focused(true); }
        public void windowDeactivated(WindowEvent e) { Focused(false); }
    public void componentMoved(ComponentEvent e) { PosChange(window.getLocation().x + insets.left, window.getLocation().y + insets.top); }

        public void componentResized(ComponentEvent e) {
            // we have to periodically do this; I don't know why
            insets = window.getInsets();
            componentResized(window.getWidth() - insets.left - insets.right, window.getHeight() - insets.top - insets.bottom);
        }

        public void componentResized(int newwidth, int newheight) {
            int oldwidth = root.width;
            int oldheight = root.height;
            SizeChange(newwidth, newheight);

            // we do this because JVMs which don't clear the background won't force repaints of these areas
            root.dirty(Math.min(oldwidth, newwidth), 0, Math.abs(oldwidth - newwidth), Math.max(oldheight, newheight));
            root.dirty(0, Math.min(oldheight, newheight), Math.max(oldwidth, newwidth), Math.abs(oldheight - newheight));

            ourGraphics = null;
        }

        public void keyTyped(KeyEvent k) { }
        public void keyPressed(KeyEvent k) { KeyPressed(translateKey(k)); }
        public void keyReleased(KeyEvent k) { KeyReleased(translateKey(k)); }
        public void mouseExited(MouseEvent m) { mouseMoved(m); }
        public void mouseEntered(MouseEvent m) { mouseMoved(m); }
        public void mouseDragged(MouseEvent m) { mouseMoved(m); }
        public void mouseMoved(MouseEvent m) {

            // ugly hack for Java1.4 dynamicLayout on Win32 -- this catches contractions during smooth resize
            int newwidth = window.getWidth() - insets.left - insets.right;
            int newheight = window.getHeight() - insets.top - insets.bottom;
            if (newwidth != root.width || newheight != root.height) componentResized(newwidth, newheight);
            
            Move(m.getX() - insets.left, m.getY() - insets.top);
        }
        public void mousePressed(MouseEvent m) { Press(modifiersToButtonNumber(m.getModifiers())); }
        public void mouseReleased(MouseEvent m) { Release(modifiersToButtonNumber(m.getModifiers())); }
        public void mouseClicked(MouseEvent m) {
            if (m.getClickCount() == 2) DoubleClick(modifiersToButtonNumber(m.getModifiers()));
            else Click(modifiersToButtonNumber(m.getModifiers()));
        }
        
        String translateKey(KeyEvent k) {
            switch (k.getKeyCode()) {
            case KeyEvent.VK_ALT: return "alt";
            case KeyEvent.VK_BACK_SPACE: return "back_space";
            case KeyEvent.VK_CONTROL: return "control";
            case KeyEvent.VK_DELETE: return "delete";
            case KeyEvent.VK_DOWN: return "down";
            case KeyEvent.VK_END: return "end";
            case KeyEvent.VK_ENTER: return "enter";
            case KeyEvent.VK_ESCAPE: return "escape";
            case KeyEvent.VK_F1: return "f1";
            case KeyEvent.VK_F10: return "f10";
            case KeyEvent.VK_F11: return "f11";
            case KeyEvent.VK_F12: return "f12";
            case KeyEvent.VK_F2: return "f2";
            case KeyEvent.VK_F3: return "f3";
            case KeyEvent.VK_F4: return "f4";
            case KeyEvent.VK_F5: return "f5";
            case KeyEvent.VK_F6: return "f6";
            case KeyEvent.VK_F7: return "f7";
            case KeyEvent.VK_F8: return "f8";
            case KeyEvent.VK_F9: return "f9";
            case KeyEvent.VK_HOME: return "home";
            case KeyEvent.VK_INSERT: return "insert";
            case KeyEvent.VK_LEFT: return "left";
            case KeyEvent.VK_META: return "alt";
            case KeyEvent.VK_PAGE_DOWN: return "page_down";
            case KeyEvent.VK_PAGE_UP: return "page_up";
            case KeyEvent.VK_PAUSE: return "pause";
            case KeyEvent.VK_PRINTSCREEN: return "printscreen";
            case KeyEvent.VK_RIGHT: return "right";
            case KeyEvent.VK_SHIFT: return "shift";
            case KeyEvent.VK_TAB: return "tab";
            case KeyEvent.VK_UP: return "up";
            default:
                char c = k.getKeyChar();
                if (c >= 1 && c <= 26) c = (char)('a' + c - 1);
                return String.valueOf(c);
            }
        }
    }

    protected ImageDecoder _decodeJPEG(InputStream is, String name) {
        try {
            Image i = Toolkit.getDefaultToolkit().createImage(org.xwt.Resources.isToByteArray(is));
            MediaTracker mediatracker = new MediaTracker(new Canvas());
            mediatracker.addImage(i, 1);
            try { mediatracker.waitForAll(); } catch (InterruptedException e) { }
            mediatracker.removeImage(i);
            final int width = i.getWidth(null);
            final int height = i.getHeight(null);
            final int[] data = new int[width * height];
            PixelGrabber pg = new PixelGrabber(i, 0, 0, width, height, data, 0, width);
            pg.grabPixels();
            if ((pg.getStatus() & ImageObserver.ABORT) != 0) {
                Log.log(this, "PixelGrabber reported an error while decoding JPEG image " + name);
                return null;
            }
            return new ImageDecoder() {
                    public int getWidth() { return width; }
                    public int getHeight() { return height; }
                    public int[] getData() { return data; }
                };
        } catch (Exception e) {
            Log.log(this, "Exception caught while decoding JPEG image " + name);
            Log.log(this, e);
            return null;
        }
    }

    // Font Handling Stuff //////////////////////////////////////////////////////////

    protected String[] _listFonts() { return fontList; }
    private static String[] fontList;
    static {
        /*
        String[] awtfonts = Toolkit.getDefaultToolkit().getFontList();
        fontList = new String[awtfonts.length * 4];
        for(int i=0; i<awtfonts.length; i++) {
            fontList[i * 4] = awtfonts[i] + "*";
            fontList[i * 4 + 1] = awtfonts[i] + "*b";
            fontList[i * 4 + 2] = awtfonts[i] + "*i";
            fontList[i * 4 + 3] = awtfonts[i] + "*bi";
        }
        */
        fontList = new String[] { };
    }

    private static Hash fontCache = new Hash();
    private static ParsedFont pf = new ParsedFont();
    private static MetricatedFont getFont(String font) {
        MetricatedFont ret = (MetricatedFont)fontCache.get(font);
        if (ret == null) {
            pf.parse(font);
            if (pf.name.equals("tty")) pf.name = "monospace";
            
            // Java's fonts tend to be, on average, two points smaller than Win32/X11 fonts. This is most acute in
            // the proxy password dialog on Linux
            ret = new MetricatedFont(pf.name, (pf.bold ? Font.BOLD : 0) | (pf.italic ? Font.ITALIC : 0), pf.size + 2);

            fontCache.put(font, ret);
        }
        return ret;
    }
    
    private static class MetricatedFont extends Font {
        public FontMetrics metrics = null;
        public MetricatedFont(String name, int size, int style) {
            super(name, size, style);
            metrics = Toolkit.getDefaultToolkit().getFontMetrics(this);
        }
    }
            
}
